"""
TEI (Text Encoding Initiative) XML processing module
"""

import datetime
import hashlib

from bs4 import BeautifulSoup
from dateutil import parser
from nltk.tokenize import sent_tokenize

from ..schema.article import Article
from ..table import Table
from ..text import Text


class TEI:
    """
    Methods to transform TEI (Text Encoding Initiative) XML into article objects.
    """

    @staticmethod
    def parse(stream, source):
        """
        Parses a TEI XML datastream and returns a processed article.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None

        Returns:
            Article
        """

        soup = BeautifulSoup(stream, "lxml")

        title = soup.find("title")
        title = title.text if title and title.text else None

        # Extract article metadata
        (
            published,
            publication,
            authors,
            affiliations,
            affiliation,
            reference,
        ) = TEI.metadata(soup)

        # Validate parsed data
        if not title and not reference:
            print("Failed to parse content - no unique identifier found")
            return None

        # Parse text sections
        sections = TEI.text(soup, title)

        # Derive uid
        uid = hashlib.sha1(title.encode("utf-8") if title else reference.encode("utf-8")).hexdigest()

        # Default title to source if empty
        title = title if title else source

        # Article metadata - id, source, published, publication, authors, affiliations, affiliation, title,
        #                    tags, reference, entry date
        metadata = (
            uid,
            source,
            published,
            publication,
            authors,
            affiliations,
            affiliation,
            title,
            "PDF",
            reference,
            parser.parse(datetime.datetime.now().strftime("%Y-%m-%d")),
        )

        # Citation references
        citations = TEI.citations(soup)

        # Create article and return
        return Article(metadata, sections, citations)

    @staticmethod
    def date(published):
        """
        Attempts to parse a publication date, if available. Otherwise, None is returned.

        Args:
            published: published object

        Returns:
            publication date if available/found, None otherwise
        """

        # Parse publication date
        # pylint: disable=W0702
        try:
            published = parser.parse(published["when"]) if published and "when" in published.attrs else None
        except:
            published = None

        return published

    @staticmethod
    def authors(source):
        """
        Parses authors and associated affiliations from the article.

        Args:
            elements: authors elements

        Returns:
            (semicolon separated list of authors, semicolon separated list of affiliations, primary affiliation)
        """

        authors = []
        affiliations = []

        for name in source.find_all("persname"):
            surname = name.find("surname")
            forename = name.find("forename")

            if surname and forename:
                authors.append(f"{surname.text}, {forename.text}")

        for affiliation in source.find_all("affiliation"):
            names = [name.text for name in affiliation.find_all("orgname")]
            affiliations.append((", ".join(names)))

        return (
            "; ".join(authors),
            "; ".join(dict.fromkeys(affiliations)),
            affiliations[-1] if affiliations else None,
        )

    @staticmethod
    def metadata(soup):
        """
        Extracts article metadata.

        Args:
            soup: bs4 handle

        Returns:
            (published, publication, authors, reference)
        """

        # Build reference link
        source = soup.find("sourcedesc")
        if source:
            published = source.find("monogr").find("date")
            publication = source.find("monogr").find("title")

            # Parse publication information
            published = TEI.date(published)
            publication = publication.text if publication else None
            authors, affiliations, affiliation = TEI.authors(source)

            struct = soup.find("biblstruct")
            reference = "https://doi.org/" + struct.find("idno").text if struct and struct.find("idno") else None
        else:
            published, publication, authors, affiliations, affiliation, reference = (
                None,
                None,
                None,
                None,
                None,
                None,
            )

        return (published, publication, authors, affiliations, affiliation, reference)

    @staticmethod
    def abstract(soup, title):
        """
        Builds a list of title and abstract sections.

        Args:
            soup: bs4 handle
            title: article title

        Returns:
            list of sections
        """

        sections = [("TITLE", title)]

        abstract = soup.find("abstract").text
        if abstract:
            # Transform and clean text
            abstract = Text.transform(abstract)
            abstract = abstract.replace("\n", " ")

            sections.extend([("ABSTRACT", x) for x in sent_tokenize(abstract)])

        return sections

    @staticmethod
    def text(soup, title):
        """
        Builds a list of text sections.

        Args:
            soup: bs4 handle
            title: article title

        Returns:
            list of sections
        """

        # Initialize with title and abstract text
        sections = TEI.abstract(soup, title)

        for section in soup.find("text").find_all("div", recursive=False):
            # Section name and text
            children = list(section.children)

            # Attempt to parse section header
            if children and not children[0].name:
                name = str(children[0]).upper()
                children = children[1:]
            else:
                name = None

            text = " ".join([str(e.text) if hasattr(e, "text") else str(e) for e in children])
            text = text.replace("\n", " ")

            # Transform and clean text
            text = Text.transform(text)

            # Split text into sentences, transform text and add to sections
            sections.extend([(name, x) for x in sent_tokenize(text)])

        # Extract text from tables
        for i, figure in enumerate(soup.find("text").find_all("figure")):
            # Use XML Id (if available) as figure name to ensure figures are uniquely named
            name = figure.get("xml:id")
            name = name.upper() if name else f"FIGURE_{i}"

            # Search for table
            table = figure.find("table")
            if table:
                sections.extend([(name, x) for x in Table.extract(table)])

        return sections

    @staticmethod
    def citations(soup):
        """
        Gets a list of citation references for this article.

        Args:
            soup: bs4 handle

        Returns:
            list of citation references
        """

        # Citation references
        citations = []

        references = soup.find("div", {"type": "references"})
        if references:
            for title in references.find("title", {"level": "a"}):
                citations.append(hashlib.sha1(title.encode("utf-8")).hexdigest())

        return citations



"""
Elasticsearch module
"""

# Conditional import
try:
    from elasticsearch import Elasticsearch, helpers

    ELASTICSEARCH = True
except ImportError:
    ELASTICSEARCH = False

from .database import Database


class Elastic(Database):
    """
    Defines data structures and methods to store article content in Elasticsearch.
    """

    # Articles index
    ARTICLES = {
        "settings": {
            "number_of_shards": 5,
            "number_of_replicas": 0,
            "index.mapping.nested_objects.limit": 30000,
        },
        "mappings": {"properties": {"sections": {"type": "nested"}}},
    }

    def __init__(self, url, replace):
        """
        Connects and initializes an elasticsearch instance.

        Args:
            url: elasticsearch url
            replace: If database should be recreated
        """

        if not ELASTICSEARCH:
            raise ImportError('Elasticsearch is not available - install "elasticsearch" extra to enable')

        # Connect to ES instance
        self.connection = Elasticsearch(hosts=[url], request_timeout=60, retry_on_timeout=True)

        # Row count
        self.rows = 0

        # Buffered actions
        self.buffer = []

        # Check if index exists
        exists = self.connection.indices.exists(index="articles")

        # Delete if replace enabled
        if exists and replace:
            self.connection.indices.delete(index="articles")
            exists = False

        # Create if necessary
        if not exists:
            self.connection.indices.create(
                index="articles",
                settings=Elastic.ARTICLES["settings"],
                mappings=Elastic.ARTICLES["mappings"],
            )

    def save(self, article):
        # Build article
        article = article.build()

        # Bulk action fields
        article["_id"] = article["id"]
        article["_index"] = "articles"

        # Buffer article
        self.buffer.append(article)

        # Increment number of articles processed
        self.rows += 1

        # Bulk load every 1000 records
        if self.rows % 1000 == 0:
            helpers.bulk(self.connection, self.buffer)
            self.buffer = []

            print(f"Inserted {self.rows} articles", end="\r")

    def complete(self):
        # Load remaining buffered articles
        if self.buffer:
            helpers.bulk(self.connection, self.buffer)

        print(f"Total articles inserted: {self.rows}")

        # Refresh indices
        self.connection.indices.refresh(index="articles")

    def close(self):
        self.connection.close()



"""
Database module
"""


class Database:
    """
    Defines data structures and methods to store article content.
    """

    def save(self, article):
        """
        Saves an article.

        Args:
            article: article metadata and text content
        """

    def complete(self):
        """
        Signals processing is complete and runs final storage methods.
        """

    def close(self):
        """
        Commits and closes the database.
        """



"""
arXiv XML processing module
"""

import hashlib
import re

from bs4 import BeautifulSoup
from dateutil import parser
from nltk.tokenize import sent_tokenize

from ..schema.article import Article
from ..text import Text


class ARX:
    """
    Methods to transform arXiv XML into article objects.
    """

    @staticmethod
    def parse(stream, source):
        """
        Parses a XML datastream and yields processed articles.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None
            config: path to config directory
        """

        # Parse XML
        soup = BeautifulSoup(stream, "lxml")

        # Process each entry
        for entry in soup.find_all("entry"):
            reference = ARX.get(entry, "id")
            title = ARX.get(entry, "title")
            published = parser.parse(ARX.get(entry, "published").split("T")[0])
            updated = parser.parse(ARX.get(entry, "updated").split("T")[0])

            # Derive uid
            uid = hashlib.sha1(reference.encode("utf-8")).hexdigest()

            # Get journal reference
            journal = ARX.get(entry, "arxiv:journal_ref")

            # Get authors
            authors, affiliations, affiliation = ARX.authors(entry.find_all("author"))

            # Get tags
            tags = "; ".join(["ARX"] + [category.get("term") for category in entry.find_all("category")])

            # Transform section text
            sections = ARX.sections(title, ARX.get(entry, "summary"))

            # Article metadata - id, source, published, publication, authors, affiliations, affiliation, title,
            #                    tags, reference, entry date
            metadata = (
                uid,
                source,
                published,
                journal,
                authors,
                affiliations,
                affiliation,
                title,
                tags,
                reference,
                updated,
            )

            yield Article(metadata, sections)

    @staticmethod
    def get(element, path):
        """
        Finds the first matching path in element and returns the element text.

        Args:
            element: XML element
            path: path expression

        Returns:
            string
        """

        element = element.find(path)
        return ARX.clean(element.text) if element else None

    @staticmethod
    def clean(text):
        """
        Removes newlines and extra spacing from text.

        Args:
            text: text to clean

        Returns:
            clean text
        """

        # Remove newlines and cleanup spacing
        text = text.replace("\n", " ")
        return re.sub(r"\s+", " ", text).strip()

    @staticmethod
    def authors(elements):
        """
        Parses authors and associated affiliations from the article.

        Args:
            elements: authors elements

        Returns:
            (semicolon separated list of authors, semicolon separated list of affiliations, primary affiliation)
        """

        authors = []
        affiliations = []

        for author in elements:
            # Create authors as lastname, firstname
            name = ARX.get(author, "name")
            authors.append(", ".join(name.rsplit(maxsplit=1)[::-1]))

            # Add affiliations
            affiliations.extend([ARX.clean(affiliation.text) for affiliation in author.find_all("arxiv:affiliation")])

        return (
            "; ".join(authors),
            "; ".join(dict.fromkeys(affiliations)),
            affiliations[-1] if affiliations else None,
        )

    @staticmethod
    def sections(title, text):
        """
        Gets a list of sections for this article.

        Args:
            title: title string
            text: summary text

        Returns:
            list of sections
        """

        # Add title
        sections = [("TITLE", title)]

        # Transform and clean text
        text = Text.transform(text)

        # Split text into sentences, transform text and add to sections
        sections.extend([("ABSTRACT", x) for x in sent_tokenize(text)])

        return sections



"""
PDF processing module
"""

from io import StringIO

import requests

from .tei import TEI


class PDF:
    """
    Methods to transform medical/scientific PDFs into article objects.
    """

    @staticmethod
    def parse(stream, source):
        """
        Parses a medical/scientific PDF datastream and returns a processed article.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None

        Returns:
            Article
        """

        # Attempt to convert PDF to TEI XML
        xml = PDF.convert(stream)

        # Parse and return object
        return TEI.parse(xml, source) if xml else None

    @staticmethod
    def convert(stream):
        """
        Converts a medical/scientific article PDF into TEI XML via a GROBID Web Service API call.

        Args:
            stream: handle to input data stream

        Returns:
            TEI XML stream
        """

        # Call GROBID API
        response = requests.post("http://localhost:8070/api/processFulltextDocument", files={"input": stream})

        # Validate request was successful
        if not response.ok:
            print(f"Failed to process file - {response.text}")
            return None

        # Wrap as StringIO
        return StringIO(response.text)



"""
Transforms and loads medical/scientific files into an articles database.
"""

import gzip
import os
import pickle
import tempfile

from multiprocessing import Process, Queue

from ..factory import Factory

from .arx import ARX
from .csvf import CSV
from .pdf import PDF
from .pmb import PMB
from .tei import TEI


class Execute:
    """
    Transforms and loads medical/scientific files into an articles database.
    """

    # Completion process signal
    COMPLETE = 1

    @staticmethod
    def mode(source, extension):
        """
        Determines file open mode for source file.

        Args:
            source: text string describing stream source
            extension: data format

        Returns:
            file open mode
        """

        return "rb" if extension == "pdf" or (source and source.lower().startswith("pubmed")) else "r"

    @staticmethod
    def parse(path, source, extension, compress, config):
        """
        Parses articles from file at path.

        Args:
            path: path to input file
            source: text string describing stream source
            extension: data format
            config: path to config directory
        """

        print(f"Processing: {path}")

        # Determine if file needs to be open in binary or text mode
        mode = Execute.mode(source, extension)

        with gzip.open(path, mode) if compress else open(path, mode, encoding="utf-8" if mode == "r" else None) as stream:
            if extension == "pdf":
                yield PDF.parse(stream, source)
            elif extension == "xml":
                if source and source.lower().startswith("arxiv"):
                    yield from ARX.parse(stream, source)
                elif source and source.lower().startswith("pubmed"):
                    yield from PMB.parse(stream, source, config)
                else:
                    yield TEI.parse(stream, source)
            elif extension == "csv":
                yield from CSV.parse(stream, source)

    @staticmethod
    def process(inputs, outputs, batchsize):
        """
        Main worker process loop. Processes file paths stored in inputs and writes articles
        to outputs. Writes a final message upon completion.

        Args:
            inputs: inputs queue
            outputs: outputs queue
            batchsize: batch size
        """

        batch = []
        try:
            # Process until inputs queue is exhausted
            while not inputs.empty():
                params = inputs.get()

                # Parse file and save successfully parsed (not None) results
                for result in Execute.parse(*params):
                    if result:
                        batch.append(result)
                        if len(batch) == batchsize:
                            outputs.put(Execute.serialize(batch))
                            batch = []

        finally:
            # Final batch
            if batch:
                outputs.put(Execute.serialize(batch))

            # Write message that process is complete
            outputs.put(Execute.COMPLETE)

    @staticmethod
    def serialize(batch):
        """
        Saves a batch of data to a temporary file for later processing.

        Args:
            batch to save

        Returns:
            temporary file name
        """

        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as output:
            # Serialize batch to temporary file
            pickle.dump(batch, output)

            # Temporary file name
            return output.name

    @staticmethod
    def scan(indir, config, inputs):
        """
        Scans for files in indir and writes to inputs queue.

        Args:
            indir: input directory
            config: path to config directory, if any
            inputs: inputs queue

        Returns:
            total number of items put into inputs queue
        """

        # Total number of files put into input queue
        total = 0

        # Recursively walk directory looking for files
        for root, _, files in sorted(os.walk(indir)):
            for f in sorted(files):
                # Extract file extension
                parts = f.lower().split(".")
                extension, compress = (parts[-2], True) if parts[-1] == "gz" else (parts[-1], False)

                # Check if file ends with accepted extension
                if any(extension for ext in ["csv", "pdf", "xml"] if ext == extension):
                    # Build full path to file
                    path = os.path.join(root, f)

                    # Write parameters to inputs queue
                    inputs.put((path, f, extension, compress, config))
                    total += 1

        return total

    @staticmethod
    def save(processes, outputs, db):
        """
        Main consumer loop that saves articles created by worker processes.

        Args:
            processes: list of worker processes
            outputs: outputs queue
            db: output database
        """

        # Read output from worker processes
        empty, complete = False, 0
        while not empty:
            # Get next result
            result = outputs.get()

            # Mark process as complete if all workers are complete and output queue is empty
            if result == Execute.COMPLETE:
                complete += 1
                empty = len(processes) == complete and outputs.empty()

            # Save article, this method will skip duplicates based on entry date
            elif result:
                with open(result, "rb") as f:
                    for x in pickle.load(f):
                        db.save(x)

                # Delete temporary file
                os.remove(result)

    @staticmethod
    def close(processes, inputs, outputs):
        """
        Closes open processes and queues.

        Args:
            processes: list of processes
            inputs: input queue
            outputs: output queue
        """

        if processes:
            # Close processes
            for process in processes:
                process.close()

            # Close queues
            inputs.close()
            outputs.close()

    @staticmethod
    def run(indir, url, config=None, replace=False, batchsize=32):
        """
        Main execution method.

        Args:
            indir: input directory
            url: database url
            config: path to config directory, if any
            replace: if true, a new database will be created, overwriting any existing database
            batchsize: batch size
        """

        processes, inputs, outputs = None, None, None
        try:
            # Build database connection
            db = Factory.create(url, replace)

            # Create queues
            inputs, outputs = Queue(), Queue()

            # Scan input directory and add files to inputs queue
            total = Execute.scan(indir, config, inputs)

            # Start worker processes
            processes = []
            for _ in range(min(total, os.cpu_count())):
                process = Process(target=Execute.process, args=(inputs, outputs, batchsize))
                process.start()
                processes.append(process)

            # Read results from worker processes and save to database
            Execute.save(processes, outputs, db)

            # Complete and close database
            db.complete()
            db.close()

            # Wait for processes to terminate
            for process in processes:
                process.join()

        finally:
            Execute.close(processes, inputs, outputs)



"""
Text module
"""

import re

# Compiled pattern for cleaning text
# pylint: disable=W0603
PATTERN = None


def getPattern():
    """
    Gets or builds a pre-compiled regex for cleaning text.

    Returns:
        compiled regex
    """

    global PATTERN

    if not PATTERN:
        # List of patterns
        patterns = []

        # Remove emails
        patterns.append(r"\w+@\w+(\.[a-z]{2,})+")

        # Remove urls
        patterns.append(r"http(s)?\:\/\/\S+")

        # Remove single characters repeated at least 3 times (ex. j o u r n a l)
        patterns.append(r"(^|\s)(\w\s+){3,}")

        # Remove citations references (ex. [3] [4] [5])
        patterns.append(r"(\[\d+\]\,?\s?){3,}(\.|\,)?")

        # Remove citations references (ex. [3, 4, 5])
        patterns.append(r"\[[\d\,\s]+\]")

        # Remove citations references (ex. (NUM1) repeated at least 3 times with whitespace
        patterns.append(r"(\(\d+\)\s){3,}")

        PATTERN = re.compile("|".join([f"({p})" for p in patterns]))

    return PATTERN


class Text:
    """
    Methods for formatting and cleaning text.
    """

    @staticmethod
    def transform(text):
        """
        Transforms and cleans text to help improve text indexing accuracy.

        Args:
            text: input text line

        Returns:
            transformed text
        """

        # Clean/transform text
        text = getPattern().sub(" ", text)

        # Remove extra spacing either caused by replacements or already in text
        text = re.sub(r" {2,}|\.{2,}", " ", text)

        return text



"""
CSV processing module
"""

import csv
import datetime

from dateutil import parser

from ..schema.article import Article


class CSV:
    """
    Methods to transform CSVs into article objects.
    """

    @staticmethod
    def parse(stream, source):
        """
        Parses a CSV datastream and yields processed articles.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None
        """

        for row in csv.DictReader(stream):
            # Parse metadata
            metadata = CSV.metadata(row, source)

            # Parse sections
            sections = CSV.sections(row)

            yield Article(metadata, sections)

    @staticmethod
    def metadata(row, source):
        """
        Parses a metadata tuple from input CSV row.

        Args:
            row: dict record with fields
            source: text string describing stream source, can be None

        Returns:
            metadata tuple
        """

        # Article metadata - id, source, published, publication, authors, affiliations, affiliation, title,
        #                    tags, reference, entry date
        fields = (
            "id",
            "source",
            "published",
            "publication",
            "authors",
            "affiliations",
            "affiliation",
            "title",
            "tags",
            "reference",
            "entry",
        )

        metadata = []
        for field in fields:
            value = None
            if field == "source":
                value = row.get(field, source)
            elif field == "entry":
                # Parse date field if found, otherwise use current date
                value = row.get(field)
                value = parser.parse(value if value else datetime.datetime.now().strftime("%Y-%m-%d"))
            else:
                value = row.get(field)

            metadata.append(value)

        return tuple(metadata)

    @staticmethod
    def sections(row):
        """
        Parses section text data from input CSV row.

        Args:
            row: dict record with fields

        Returns:
            list of sections
        """

        # Start with title as text
        text = row.get("title")

        # Append abstract if available
        abstract = row.get("abstract")
        if abstract:
            text += " " + abstract

        # Create single section from text
        return [(None, text)]



"""
Factory module
"""

from .elastic import Elastic
from .filesystem import JSON, YAML
from .sqlite import SQLite


class Factory:
    """
    Database Factory - creates database connections
    """

    @staticmethod
    def create(url, replace):
        """
        Creates a new database connection.

        Args:
            url: connection url
            replace: if true, a new database will be created, overwriting any existing database

        Returns:
            Database
        """

        if url.startswith("http://"):
            return Elastic(url, replace)
        if url.startswith("json://"):
            return JSON(url.replace("json://", ""))
        if url.startswith("yaml://"):
            return YAML(url.replace("yaml://", ""))
        if url:
            # If URL is present, assume it's SQLite
            return SQLite(url.replace("sqlite://", ""), replace)

        return None



"""
PubMed archive XML processing module
"""

import os
import re

from dateutil import parser
from lxml import etree
from nltk.tokenize import sent_tokenize

from ..schema.article import Article
from ..text import Text


class PMB:
    """
    Methods to transform PubMed archive XML files into article objects.
    """

    @staticmethod
    def load(config, name):
        """
        Loads a set of filters from a configuration file.

        Args:
            config: config directory path
            name: name of config file

        Returns:
            set of filters
        """

        filters = None

        path = os.path.join(config, name) if config else None
        if path and os.path.exists(path):
            with open(path, encoding="utf-8") as f:
                filters = set(line.strip() for line in f)

        return filters

    # pylint: disable=W0613
    @staticmethod
    def parse(stream, source, config):
        """
        Parses a XML datastream and yields processed articles.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None
            config: path to config directory
        """

        # Load id, MeSH and keyword filters, if available
        ids, codes, keywords = PMB.load(config, "ids"), PMB.load(config, "codes"), PMB.load(config, "keywords")

        # Convert ids to ints
        ids = set(int(x) for x in ids) if ids else None

        # Parse HTML content using lxml
        # pylint: disable=c-extension-no-member,stop-iteration-return
        document = etree.iterparse(stream, events=("start", "end"))
        _, root = next(document)

        for event, element in document:
            if event == "end" and element.tag == "PubmedArticle":
                yield PMB.process(element, source, ids, codes, keywords)
                root.clear()

    @staticmethod
    def process(element, source, ids, codes, keywords):
        """
        Processes a single XML article element into an Article.

        This method applies the following logic for filters.

          - Match on ids
          - Match on MeSH codes
          - Match on keywords

        Filters only fail if there is a list of terms to check. When there are multiple filters active, only one
        filter has to pass.

        Args:
            element: XML element
            source: text string describing stream source, can be None
            ids: List of ids to select, can be None
            codes: List of MeSH codes to select, can be None
            keywords: List of keywords to search for, can be None

        Returns:
            Article or None if Article not parsed
        """

        citation = element.find("MedlineCitation")
        article = citation.find("Article")
        journal = article.find("Journal")

        # General fields
        uid = int(citation.find("PMID").text)

        # Apply ids filter, skip article if ids filter fails and no other filters active
        idsfail = ids and uid not in ids
        if idsfail and not codes and not keywords:
            return None

        # Apply MeSH codes filter, skip article if all active filters fail
        mesh = PMB.mesh(citation)
        codesfail = codes and not any(x for x in mesh if x in codes)
        if (not ids or idsfail) and codesfail and not keywords:
            return None

        source = source if source else "PMB"
        reference = f"https://pubmed.ncbi.nlm.nih.gov/{uid}"
        entry = PMB.date(citation.find("DateRevised"))

        # Journal fields
        published = PMB.published(journal)
        publication = PMB.get(journal, "Title")

        # Article fields
        title = PMB.get(article, "ArticleTitle")
        authors, affiliations, affiliation = PMB.authors(article)

        # Create tags
        tags = "; ".join(["PMB"] + mesh)

        # Abstract text
        sections = PMB.sections(article, title)

        # Citation references
        citations = PMB.citations(element, ids)

        # Apply keywords filter, skip article if all active filters fail
        wordsfail = keywords and not any(k for k in keywords if any(x for _, x in sections if k.lower() in x.lower()))
        if (not ids or idsfail) and (not codes or codesfail) and wordsfail:
            return None

        # Require title (title is in sections) and at least one other section
        if len(sections) > 1:
            metadata = (str(uid), source, published, publication, authors, affiliations, affiliation, title, tags, reference, entry)
            return Article(metadata, sections, citations)

        return None

    @staticmethod
    def get(element, path):
        """
        Finds the first matching path in element and returns the element text.

        Args:
            element: XML element
            path: path expression

        Returns:
            string
        """

        element = element.find(path)
        return PMB.text(element)

    @staticmethod
    def text(element):
        """
        Flattens elements into a single text string.

        Args:
            element: XML element

        Returns:
            string
        """

        return "".join(element.itertext()) if element is not None else None

    @staticmethod
    def date(element):
        """
        Attempts to parse a date from an element.

        Args:
            element: input element

        Return:
            Date if parsed
        """

        date = ""
        for field in ["Year", "Month", "Day"]:
            value = PMB.get(element, field)
            if value:
                date += "-" + value if date else value

        return parser.parse(date) if date else None

    @staticmethod
    def published(journal):
        """
        Parses the published date. Multiple date formats are handled via the
        dateparser library.

        Args:
            journal: journal element

        Returns:
            datetime
        """

        element = journal.find("JournalIssue/PubDate")

        date = PMB.date(element)
        if not date:
            # Fallback to MedlineDate
            date = PMB.get(element, "MedlineDate")
            date = re.search(r"\d{4}", date)
            date = parser.parse(date.group()) if date else None

        return date if date else None

    @staticmethod
    def authors(journal):
        """
        Parses authors and associated affiliations from the article.

        Args:
            journal: journal element

        Returns:
            (semicolon separated list of authors, semicolon separated list of affiliations, primary affiliation)
        """

        authors = []
        affiliations = []

        for author in journal.findall("AuthorList/Author"):
            lastname = PMB.get(author, "LastName")
            forename = PMB.get(author, "ForeName")

            # Add author affiliations
            for affiliation in author.findall("AffiliationInfo/Affiliation"):
                affiliations.append(PMB.text(affiliation))

            if lastname and forename:
                authors.append(f"{lastname}, {forename}")

        return (
            "; ".join(authors),
            "; ".join(dict.fromkeys(affiliations)),
            affiliations[-1] if affiliations else None,
        )

    @staticmethod
    def mesh(citation):
        """
        Gets a list of article MeSH codes.

        Args:
            citation: citation element

        Returns:
            list of MeSH codes
        """

        return [descriptor.attrib["UI"] for descriptor in citation.findall("MeshHeadingList//DescriptorName") if descriptor.attrib["UI"]]

    @staticmethod
    def sections(article, title):
        """
        Gets a list of sections for this article. This method supports the following three abstract formats:
           - Raw text
           - HTML formatted text
           - Abstract text parsed into section named elements

        All three formats return sections that are tokenized into sentences.

        Args:
            article: article element
            title: title string

        Returns:
            list of sections
        """

        sections = [("TITLE", title)] if title else []

        # Get AbstractText elements
        elements = article.findall("Abstract/AbstractText")

        if len(elements) == 1:
            element = elements[0]
            if element.text:
                # Process abstract as raw text
                sections.extend(PMB.raw(element))
            else:
                # Process abstract as formatted text
                sections.extend(PMB.formatted(element))
        else:
            # Process abstract as parsed elements with section names
            sections.extend(PMB.parsed(elements))

        return sections

    @staticmethod
    def raw(element):
        """
        Parses a raw text abstract into a list of sections.

        Args:
            element: abstract text element

        Returns:
            list of sections
        """

        # Transform and clean text
        text = Text.transform(PMB.text(element))

        # No embedded sections
        return [("ABSTRACT", x) for x in sent_tokenize(text)]

    @staticmethod
    def formatted(element):
        """
        Parses a HTML formatted text abstract into a list of sections.

        Args:
            element: abstract text element

        Returns:
            list of sections
        """

        sections = []
        name, tag, texts = "ABSTRACT", None, []

        for x in element.getchildren():
            # Get raw tag text and clean tag text
            rtext = "".join(x.itertext())
            ctext = PMB.section(rtext)
            new = False

            # Set section tag type if all of the following:
            #   - tag not set
            #   - cleaned inner text has data
            #   - no section text queued
            #   - element tag is a <b> or matches a defined section background category name
            if not tag and ctext and not texts and (x.tag.lower() == "b" or PMB.background(ctext)):
                tag = x.tag

            # New section if one of following:
            #   - element tag matches section tag and tag text has data
            #   - no section tag and section text is queued
            # AND
            #   - no section text
            #   - last section text element ends in period
            # pylint: disable=R0916
            if ((x.tag == tag and ctext) or (not tag and texts)) and (not texts or texts[-1].strip().endswith(".")):
                # Save previous section
                if texts:
                    sections.extend([(name, t) for t in sent_tokenize("".join(texts).strip())])

                # Reset section name/texts
                name = ctext if tag else "ABSTRACT"
                texts = []
                new = True

            # Create section text. Skip tag text for section names.
            text = rtext if not new and ctext else ""
            text += x.tail if x.tail else ""
            if text.strip():
                texts.append(text)

        # Save last section
        if texts:
            sections.extend([(name, t) for t in sent_tokenize("".join(texts).strip())])

        return sections

    @staticmethod
    def parsed(elements):
        """
        Parses a labeled text abstract into a list of sections.

        Args:
            element: abstract text element

        Returns:
            list of sections
        """

        sections = []

        # Parsed abstract
        for element in elements:
            name = PMB.section(element.attrib["Label"]) if "Label" in element.attrib else None
            name = name if name else "ABSTRACT"

            if element.text:
                # Transform and clean text
                text = Text.transform(PMB.text(element))

                # Split text into sentences, transform text and add to sections
                sections.extend([(name, x) for x in sent_tokenize(text)])

        return sections

    @staticmethod
    def background(name):
        """
        Checks if the input section name is a background category.

        Args:
            name: section name

        Returns:
            True if the section name is a background category
        """

        return [x for x in ["aim", "introduction", "background", "purpose", "objective"] if x in name.lower()]

    @staticmethod
    def section(name):
        """
        Formats a section name.

        Args:
            name: input section name

        Returns:
            formatted section name
        """

        return re.sub(r"[^\w)]+$", "", name.strip()).upper()

    @staticmethod
    def citations(element, ids):
        """
        Gets a list of citation references for this article.

        Args:
            element: XML element
            ids: List of ids to select, can be None

        Returns:
            list of citation references
        """

        # Citation references
        citations = []

        # Add pubmed references, support id filters
        for reference in element.findall("PubmedData/ReferenceList/Reference"):
            for article in reference.findall("ArticleIdList/ArticleId"):
                # Get PubMed Id, if applicable
                uid = int(article.text) if article.text and article.text.isdigit() and article.attrib["IdType"] == "pubmed" else None

                # Validate uid is valid and in list of valid ids, if applicable
                if uid is not None and (not ids or uid in ids):
                    citations.append(article.text)

        return citations



"""
SQLite module
"""

import os
import sqlite3

from dateutil import parser

from .database import Database


class SQLite(Database):
    """
    Defines data structures and methods to store article content in SQLite.
    """

    # Articles schema
    ARTICLE = (
        "id",
        "source",
        "published",
        "publication",
        "authors",
        "affiliations",
        "affiliation",
        "title",
        "tags",
        "reference",
        "entry",
    )

    # Articles schema
    ARTICLES = {
        "Id": "TEXT PRIMARY KEY",
        "Source": "TEXT",
        "Published": "DATETIME",
        "Publication": "TEXT",
        "Authors": "TEXT",
        "Affiliations": "TEXT",
        "Affiliation": "TEXT",
        "Title": "TEXT",
        "Tags": "TEXT",
        "Reference": "TEXT",
        "Entry": "DATETIME",
    }

    # Sections schema
    SECTIONS = {
        "Id": "INTEGER PRIMARY KEY",
        "Article": "TEXT",
        "Name": "TEXT",
        "Text": "TEXT",
    }

    # Citations schema
    CITATIONS = {
        "Id": "INTEGER PRIMARY KEY",
        "Article": "TEXT",
        "Reference": "TEXT",
    }

    # SQL statements
    CREATE_TABLE = "CREATE TABLE IF NOT EXISTS {table} ({fields})"
    INSERT_ROW = "INSERT INTO {table} ({columns}) VALUES ({values})"
    CREATE_SECTION_INDEX = "CREATE INDEX section_article ON sections(article)"
    CREATE_CITATION_INDEX_1 = "CREATE INDEX citation_article ON citations(article)"
    CREATE_CITATION_INDEX_2 = "CREATE INDEX citation_reference ON citations(reference)"

    # Restore index when updating an existing database
    SECTION_COUNT = "SELECT MAX(Id) FROM sections"
    CITATION_COUNT = "SELECT MAX(Id) FROM citations"

    # Lookup entry date for an article
    LOOKUP_ENTRY = "SELECT Entry FROM articles WHERE id = ?"

    # Delete article
    DELETE_ARTICLE = "DELETE FROM articles WHERE id = ?"
    DELETE_SECTIONS = "DELETE FROM sections WHERE article = ?"

    def __init__(self, outdir, replace):
        """
        Creates and initializes a new output SQLite database.

        Args:
            outdir: output directory
            replace: If database should be recreated
        """

        # Create if output path doesn't exist
        os.makedirs(outdir, exist_ok=True)

        # Output database file
        dbfile = os.path.join(outdir, "articles.sqlite")

        # Create flag
        create = replace or not os.path.exists(dbfile)

        # Delete existing file if replace set
        if replace and os.path.exists(dbfile):
            os.remove(dbfile)

        # Index fields
        self.aindex, self.sindex, self.cindex = 0, 0, 0

        # Connect to output database
        self.db = sqlite3.connect(dbfile)

        # Create database cursor
        self.cur = self.db.cursor()

        if create:
            # Create articles table
            self.create(SQLite.ARTICLES, "articles")

            # Create sections table
            self.create(SQLite.SECTIONS, "sections")

            # Create citations table
            self.create(SQLite.CITATIONS, "citations")

            # Create indexes
            for index in [SQLite.CREATE_SECTION_INDEX, SQLite.CREATE_CITATION_INDEX_1, SQLite.CREATE_CITATION_INDEX_2]:
                self.execute(index)
        else:
            # Restore section and citation index id
            self.sindex = int(self.cur.execute(SQLite.SECTION_COUNT).fetchone()[0]) + 1
            self.cindex = int(self.cur.execute(SQLite.CITATION_COUNT).fetchone()[0]) + 1

        # Start transaction
        self.cur.execute("BEGIN")

    def save(self, article):
        # Save article if not a duplicate
        if self.savearticle(article):
            # Increment number of articles processed
            self.aindex += 1
            if self.aindex % 1000 == 0:
                print(f"Inserted {self.aindex} articles", end="\r")

                # Commit current transaction and start a new one
                self.transaction()

            for name, text in article.sections:
                # Section row - id, article, name, text
                self.insert(
                    SQLite.SECTIONS,
                    "sections",
                    (self.sindex, article.uid(), name, text),
                )
                self.sindex += 1

            for reference in article.citations:
                # Citation row - id, article, reference
                self.insert(
                    SQLite.CITATIONS,
                    "citations",
                    (self.cindex, article.uid(), reference),
                )
                self.cindex += 1

    def savearticle(self, article):
        """
        Saves an article to SQLite. If a duplicate entry is found, this method compares the entry
        date and keeps the article with the latest entry date.

        Args:
            article: article metadata and text content

        Returns
            True if article saved, False otherwise
        """

        try:
            # Article row
            self.insert(SQLite.ARTICLES, "articles", article.metadata)
        except sqlite3.IntegrityError:
            # Duplicate detected get entry date to determine action
            entry = parser.parse(self.cur.execute(SQLite.LOOKUP_ENTRY, [article.uid()]).fetchone()[0])

            # Keep existing article if existing entry date is same or newer
            if article.entry() <= entry:
                return False

            # Delete and re-insert article
            self.cur.execute(SQLite.DELETE_ARTICLE, [article.uid()])
            self.cur.execute(SQLite.DELETE_SECTIONS, [article.uid()])
            self.insert(SQLite.ARTICLES, "articles", article.metadata)

        return True

    def complete(self):
        print(f"Total articles inserted: {self.aindex}")

    def close(self):
        self.db.commit()
        self.db.close()

    def transaction(self):
        """
        Commits current transaction and creates a new one.
        """

        self.db.commit()
        self.cur.execute("BEGIN")

    def create(self, table, name):
        """
        Creates a SQLite table.

        Args:
            table: table schema
            name: table name
        """

        columns = [f"{name} {ctype}" for name, ctype in table.items()]
        create = SQLite.CREATE_TABLE.format(table=name, fields=", ".join(columns))

        # pylint: disable=W0703
        self.cur.execute(create)

    def execute(self, sql):
        """
        Executes SQL statement against open cursor.

        Args:
            sql: SQL statement
        """

        self.cur.execute(sql)

    def insert(self, table, name, row):
        """
        Builds and inserts a row.

        Args:
            table: table object
            name: table name
            row: row to insert
        """

        # Build insert prepared statement
        columns = [name for name, _ in table.items()]
        insert = SQLite.INSERT_ROW.format(table=name, columns=", ".join(columns), values=("?, " * len(columns))[:-2])

        # Execute insert statement
        self.cur.execute(insert, self.values(table, row, columns))

    def values(self, table, row, columns):
        """
        Formats and converts row into database types based on table schema.

        Args:
            table: table schema
            row: row tuple
            columns: column names

        Returns:
            Database schema formatted row tuple
        """

        values = []
        for x, column in enumerate(columns):
            # Get value
            value = row[x]

            if table[column].startswith("INTEGER"):
                values.append(int(value) if value else 0)
            elif table[column].startswith("BOOLEAN"):
                values.append(1 if value == "TRUE" else 0)
            elif table[column].startswith("TEXT"):
                # Clean empty text and replace with None
                values.append(value if value and len(value.strip()) > 0 else None)
            else:
                values.append(value)

        return values



# pylint: disable = C0111
from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    # Remove GitHub dark mode images
    DESCRIPTION = "".join([line for line in f if "gh-dark-mode-only" not in line])

# Optional dependencies
extras = {}

extras["elasticsearch"] = ["elasticsearch>=7.8.0"]
extras["all"] = extras["elasticsearch"]

setup(
    name="paperetl",
    version="2.5.2",
    author="NeuML",
    description="ETL processes for medical and scientific papers",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/neuml/paperetl",
    project_urls={
        "Documentation": "https://github.com/neuml/paperetl",
        "Issue Tracker": "https://github.com/neuml/paperetl/issues",
        "Source Code": "https://github.com/neuml/paperetl",
    },
    license="Apache 2.0: http://www.apache.org/licenses/LICENSE-2.0",
    packages=find_packages(where="src/python"),
    package_dir={"": "src/python"},
    keywords="etl parse medical scientific papers",
    python_requires=">=3.10",
    install_requires=[
        "beautifulsoup4>=4.8.1",
        "lxml>=4.5.0",
        "nltk>=3.5",
        "pandas>=1.1.0",
        "python-dateutil>=2.8.1",
        "PyYAML>=5.3",
        "requests>=2.22.0",
    ],
    extras_require=extras,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
)



"""
Defines main entry point for ETL process.
"""

import sys

from .execute import Execute

if __name__ == "__main__":
    if len(sys.argv) > 2:
        Execute.run(
            sys.argv[1],
            sys.argv[2],
            sys.argv[3] if len(sys.argv) > 3 else None,
            sys.argv[4] == "True" if len(sys.argv) > 4 else False,
            int(sys.argv[5]) if len(sys.argv) > 5 else 32,
        )



"""
Article module.
"""


class Article:
    """
    Article objects. Holds all metadata and text content related to an article.
    """

    # Articles schema
    ARTICLE = (
        "id",
        "source",
        "published",
        "publication",
        "authors",
        "affiliations",
        "affiliation",
        "title",
        "tags",
        "reference",
        "entry",
    )

    # Sections schema
    SECTION = ("name", "text")

    # Citations schema
    CITATION = ("reference",)

    def __init__(self, metadata, sections, citations=None):
        """
        Stores article metadata and section content as an object.

        Args:
            metadata: article metadata
            sections: text sections
            citations: citations
        """

        self.metadata = metadata
        self.sections = sections
        self.citations = citations if citations else []

    def uid(self):
        """
        Returns the article uid.

        Returns:
            article uid
        """

        return self.metadata[0]

    def source(self):
        """
        Returns the article source.

        Returns:
            article source
        """

        return self.metadata[1]

    def tags(self):
        """
        Returns the article tags.

        Returns:
            article tags
        """

        return self.metadata[8]

    def entry(self):
        """
        Returns the article entry date.

        Returns:
            article entry date
        """

        return self.metadata[10]

    def build(self):
        """
        Builds an article with all metadata and section content.

        Returns:
            object with full article content
        """

        # Create article
        article = dict(zip(Article.ARTICLE, self.metadata))

        # Create sections
        sections = [dict(zip(Article.SECTION, section)) for section in self.sections]

        # Create citations
        citations = [dict(zip(Article.CITATION, citation)) for citation in self.citations]

        # Add sections and citations to article
        article["sections"] = sections
        article["citations"] = citations

        return article
